module ft8_decode

  parameter (MAXFOX=1000)
  character*12 c2fox(MAXFOX)
  character*4  g2fox(MAXFOX)
  integer nsnrfox(MAXFOX)
  integer nfreqfox(MAXFOX)
  integer n30fox(MAXFOX)
  integer n30z
  integer nfox
  
  type :: ft8_decoder
     procedure(ft8_decode_callback), pointer :: callback
   contains
     procedure :: decode
  end type ft8_decoder

  abstract interface
     subroutine ft8_decode_callback (this,sync,snr,dt,freq,decoded,nap,qual)
       import ft8_decoder
       implicit none
       class(ft8_decoder), intent(inout) :: this
       real, intent(in) :: sync
       integer, intent(in) :: snr
       real, intent(in) :: dt
       real, intent(in) :: freq
       character(len=37), intent(in) :: decoded
       integer, intent(in) :: nap 
       real, intent(in) :: qual 
     end subroutine ft8_decode_callback
  end interface

contains

  subroutine decode(this,callback,iwave,nQSOProgress,nfqso,nftx,newdat,  &
       nutc,nfa,nfb,ndepth,ncontest,nagain,lft8apon,lapcqonly, &
       napwid,mycall12,hiscall12,hisgrid6)
!    use wavhdr
    use timer_module, only: timer
    use prog_args                       !shm_key, exe_dir, data_dir, back by s52d
    include 'ft8/ft8_params.f90'
!    type(hdr) h

    class(ft8_decoder), intent(inout) :: this
    procedure(ft8_decode_callback) :: callback
    parameter (MAXCAND=300)
    real s(NH1,NHSYM)
    real sbase(NH1)
    real candidate(3,MAXCAND)
    real dd(15*12000)
    logical, intent(in) :: lft8apon,lapcqonly,nagain
    logical newdat,lsubtract,ldupe
    character*12 mycall12,hiscall12,mycall12_0
    character*6 hisgrid6
    integer*2 iwave(15*12000)
    integer apsym2(58)
    character datetime*13,msg37*37
!   character message*22
    character*37 allmessages(100)
    integer allsnrs(100)
    data mycall12_0/'dummy'/
    save s,dd,mycall12_0


! s52d diversity data
    real ddd(15*12000,5) !s52d: save 4 periods + data read from disk
    save ddd        
! dttable selected so that any tone get a chance for 60 degrees
    integer*2 ::  dttable7(7) = (/ 0, 10, 17, 20, 25, 30, 39 /) 
    integer*2 ::  dttable4(7) = (/ 0, 4, 7, 10, 0, 0, 0 /) 
    integer*2 ::  dttable3(7) = (/ 0, 4, 8, 0, 0, 0, 0 /) 
    integer*2  dttablet(7), dttables(7)
    integer*2 dtoffset
    integer*2 :: tdlen =7,sdlen =0  
    logical :: divsav,divread
! diversity configuration, read file once
    logical :: tdivok = .true. , fswriteok = .false. ,fdivok = .false., sdivok = .false., first = .true.
    character*64 divwdir, divrdir
    save tdivok, fswriteok, fdivok, sdivok, first, divwdir, divrdir, dttablet, dttables, tdlen, sdlen

! read diversity data    
    if (first) then
      first=.false.
      divread=.false.
      ipnow=0 ! to make compiler happy
      ipold=2 
      dttablet=dttable7  ! default on
      dttables=dttable7
! read configuration file, if it exists      
      open(100,file=trim(data_dir)//'/wsdiv.txt',status='old',err=299)
      read(100,*, err=299) i !ft4
      read(100,*, err=299) i
      if (i.eq.0) tdivok = .false. !default ON
      if (i.ne.0) then
         tdlen=i ! poblem with wrong number
         if (i.eq.3) dttablet=dttable3
         if (i.eq.4) dttablet=dttable3
         if (i.eq.7) dttablet=dttable7
      endif
        
      read(100,*, err=299) i !ft4
      read(100,*, err=299) i
      if (i.ne.0) fswriteok = .true. !default OFF
      read(100,*, err=299) i !ft4
      read(100,*, err=299) i
      if  (i.ne.0) fdivok = .true.
      read(100,*, err=299) i !ft4
      read(100,*, err=299) i
      if (i.ne.0) then 
         sdivok = .true.
         sdlen=i
         if (i.eq.3) dttables=dttable3
         if (i.eq.4) dttables=dttable4
         if (i.eq.7) dttables=dttable7
         fdivok = .true. 
         tdivok = .true. ! we need some delay before file can be read
      endif
      read(100,*, err=299) i !spare
      read(100,*, err=299) i !spare
      read(100,'(A)', err=299) divwdir
      read(100,'(A)', err=299) divrdir
!        print *,tdivok,fswriteok,fdivok,sdivok, trim(divwdir),',',trim(divrdir),'.'
      close(100)
299  continue
    endif
        




    if(mycall12.ne.mycall12_0) then
       mycall12_0=mycall12
    endif

    this%callback => callback
    write(datetime,1001) nutc        !### TEMPORARY ###
1001 format("000000_",i6.6)

    call ft8apset(mycall12,hiscall12,apsym2)
    dd=iwave
    ndecodes=0
    allmessages='                                     '
    allsnrs=0
    ifa=nfa
    ifb=nfb
    if(nagain) then
       ifa=nfqso-10
       ifb=nfqso+10
    endif

! save data for space diversity., s52d
  divsav=fswriteok
  if (fswriteok) then
    open(101,file=trim(divwdir)//datetime//'.dd',access='stream',status='REPLACE',iostat=ioerr)
    if (ioerr .gt. 0) divsav=.false. ! directory does not exist. Some more actions needed
  endif    




! For now:
! ndepth=1: no subtraction, 1 pass, belief propagation only
! ndepth=2: subtraction, 3 passes, belief propagation only
! ndepth=3: subtraction, 3 passes, bp+osd
    if(ndepth.eq.1) npass=1
    if(ndepth.ge.2) npass=3

    do ipass=1,28 !used to be 1,npass. 10..17 time diversity, 20 frame div, 21-17 space div

!    do ipass=1,npass
      newdat=.true.  ! Is this a problem? I hijacked newdat.
      syncmin=1.5
      if(ipass.eq.1) then
        lsubtract=.true.
        if(ndepth.eq.1) lsubtract=.false.
      elseif(ipass.eq.2) then
        n2=ndecodes
        if(ndecodes.eq.0) cycle
        lsubtract=.true.
      elseif(ipass.eq.3) then
        if((ndecodes-n2).eq.0) cycle
        lsubtract=.false. 
      endif 
! s52d: diversity code
      if (divsav .or. tdivok .or. sdivok) lsubtract=.true. ! it will be processed again

      if (ipass .gt. npass .and. ipass .lt. 10 ) cycle 

      if (ipass .eq. 10 ) then ! prepare time diversity and save data for space diversity
         if (divsav) then ! save data for space diversity
           sync=0.0 ! mark end of the list
           nsnr=0.0
           f1=0.0 ! it would be nice to add actual RX QRG
           write(101) sync,nsnr,xdt,f1,msg37,iaptype,qual
           write(101) dd
           close(101) ! close fast, so other task can read after time diversity
         endif
          
         call timer('timed   ',0)
         ipnow=(mod(nutc,100)/15) +1 ! 1..4, which 15s period in the minute
         ipold=ipnow -2
         if (ipold .le. 0) ipold=ipold+4 !previous odd/even period
         ddd(:,ipnow)=dd
         lsubtract=.false.
         divread=.true. ! will be false if read fails
         cycle
     endif 
! time diversity: add what is left in dd with previous odd/even period 
      if (ipass .gt. 10 .and. ipass .le. (10+tdlen) ) then
        if (.not.tdivok) cycle 
        dtoffset = dttablet(ipass-10)
        dd(1:15*12000-dtoffset) = ddd(1:15*12000-dtoffset,ipnow) +ddd(1+dtoffset:15*12000,ipold)
      endif 

      if ((ipass.gt. (10+tdlen)).and.(ipass .lt. 20 )) cycle 
! prepare data for sace diversity      
      if (ipass.eq.20) then
          call timer('timed   ',1)
          call timer('spaced   ',0)
! read data for space diversity. 
          divread = fdivok
          if (fdivok) then
           open(102,file=trim(divrdir)//datetime//'.dd',access='stream',status='old',iostat=ioerr, err=199)
           if (ioerr .eq. 0) then 
             divread = .false.
             do
              read(102,IOSTAT=ioerr,err=199) sync,nsnr,xdt,f1,msg37,iaptype,qual
              if (ioerr.ne.0) goto 199
              if ((sync .eq. 0.0) .and. (nsnr .eq. 0.0)) exit 
              ldupe=.false.
              do id=1,ndecodes
               if (msg37.eq.allmessages(id)) ldupe=.true.
              enddo
              if(.not.ldupe) then
                ndecodes=ndecodes+1
                allmessages(ndecodes)=msg37
                allsnrs(ndecodes)=nsnr
              endif
              if(.not.ldupe .and. associated(this%callback)) then
                iaptype=iaptype+300  ! frame diversity
                call this%callback(sync,nsnr,xdt,f1,msg37,iaptype,qual)
              endif
            enddo
! now read residual dd from another jt9            
           if (sdivok) then 
              read(102,IOSTAT=ioerr,err=199) ddd(:,5)
              if (ioerr.ne.0) then
                divread=.false.;
                goto 198
              endif
! s52d: scaling do not give ME advantage, IC-7610 has good AGC
!              scale1=sum(abs(ddd(4*12000:5*12000,ipnow))) ! dd saved before time diversity
!              scale2=sum(abs(ddd(4*12000:5*12000,5))) ! other wsjtx dd
!              if (scale2 .lt. 1.0) scale2 = scale1 ! unlikely to be zero, if both are 0 then it is ok to fail ;-)
!              scale2=scale1/scale2
!              print *,scale2
!               ddd(:,5)=ddd(:,5)*scale2 ! same amplitude, RX might be set differently
              divread=.true. ! proceed with space diversity
            endif 
198         close(102, status='delete',err=199) ! erase file when data is read
           endif
199        cycle ! error: probably nothing to read.         
         endif
        endif !pass=20
! passes 21 to 27 are for space diversity
      if (ipass .ge. 21 .and. ipass .le. (20+sdlen) ) then
        if (divread) then
          dtoffset = dttables(ipass-20)
          dd(1:15*12000-dtoffset) = ddd(1:15*12000-dtoffset,ipnow) +ddd(1+dtoffset:15*12000,5)
        else 
          cycle
        endif
      endif !ipass 20..27

      if (ipass .eq. 28) call timer('spaced   ',1) 
      if (ipass .gt. (20+sdlen)) cycle

! original code          




      call timer('sync8   ',0)
      maxc=MAXCAND
      call sync8(dd,ifa,ifb,syncmin,nfqso,maxc,s,candidate,   &
           ncand,sbase)
      call timer('sync8   ',1)
      do icand=1,ncand
        sync=candidate(3,icand)
        f1=candidate(1,icand)
        xdt=candidate(2,icand)
        xbase=10.0**(0.1*(sbase(nint(f1/3.125))-40.0))
        nsnr0=min(99,nint(10.0*log10(sync) - 25.5))    !### empirical ###
        call timer('ft8b    ',0)
        call ft8b(dd,newdat,nQSOProgress,nfqso,nftx,ndepth,lft8apon,      &
             lapcqonly,napwid,lsubtract,nagain,ncontest,iaptype,mycall12,   &
             hiscall12,sync,f1,xdt,xbase,apsym2,nharderrors,dmin,           &
             nbadcrc,iappass,iera,msg37,xsnr)
        call timer('ft8b    ',1)
        nsnr=nint(xsnr) 
        xdt=xdt-0.5
        hd=nharderrors+dmin
        if(nbadcrc.eq.0) then
           ldupe=.false.
           do id=1,ndecodes
! for diversity keep first decode, not strongest. There are too much duplicates. s52d
             if((msg37.eq.allmessages(id)).and.((nsnr.le.allsnrs(id)).or.(ipass.ge.10))) ldupe=.true.
!              if(msg37.eq.allmessages(id).and.nsnr.le.allsnrs(id)) ldupe=.true.
           enddo
           if(.not.ldupe) then
              ndecodes=ndecodes+1
              allmessages(ndecodes)=msg37
              allsnrs(ndecodes)=nsnr
           endif
! fort.81 file: 5+2 lines are uncommented, report only diversity. s52d
!           if (ipass .gt. 10) then
                write(81,1004) nutc,ncand,icand,ipass,iaptype,iappass,        &
                nharderrors,dmin,hd,min(sync,999.0),nint(xsnr),          &
                xdt-0.5,f1,msg37
1004            format(i6.6,2i4,i3,2i2,i3,3f6.1,i4,f11.7,f9.2,2x,a37)
                flush(81)
!           endif

           if (ipass .gt. 10) then ! pass info if t/f/s shall be aded to msg37
               if (ipass .lt. 20) then
                iaptype=iaptype+100 ! time diversity
               else
                iaptype=iaptype+200 ! space diversity
               endif
           endif


           if(.not.ldupe .and. associated(this%callback)) then
              qual=1.0-(nharderrors+dmin)/60.0 ! scale qual to [0.0,1.0]
              if (divsav .and. (ipass.lt.10)) then ! frame diversity, s52d
                write(101) sync,nsnr,xdt,f1,msg37,iaptype,qual
              endif

              call this%callback(sync,nsnr,xdt,f1,msg37,iaptype,qual)
           endif
        endif
      enddo
  enddo
  return
  end subroutine decode

end module ft8_decode
